'' =================================================================================================
''
''   File....... jm_ac_dimmer.spin
''   Purpose.... AC dimming using phase-angle modulation from zero-cross
''   Author..... Jon "JonnyMac" McPhalen 
''               Copyright (c) 2010-2015 Jon McPhalen
''               -- see below for terms of use
''   E-mail..... 
''   Started.... 23 JAN 2010
''   Updated.... 24 AUG 2015
''
'' =================================================================================================


con { fixed io pins }

  RX1 = 31                                                       ' programming / debug port
  TX1 = 30
  
  SDA = 29                                                       ' eeprom / i2c
  SCL = 28


con

  DB_CYCLES = 3                                                  ' debounce cycles

  CHANNELS = 4
  LAST_CH  = CHANNELS - 1
  

var

  long  cog                                                      ' cog running driver

  long  linefreq
                                                                  
  byte  brightness[CHANNELS]                                     ' dimmer levels, 0 to 255
  

pub start(ch0, zc)

'' Start 4-channel AC dimmer
'' -- user responsible for correct pin assignments
''    * no error checking!
'' -- assumes ch0..ch3 contiguous and in ascending order

  return startx(ch0+0, ch0+1, ch0+2, ch0+3, zc)
  

pub startx(ch0, ch1, ch2, ch3, zc) | zctix, fr

'' Start 4-channel AC dimmer
'' -- user responsible for correct pin assignments
''    * no error checking!

  stop                                                           ' stop if running

  ' setup cog before launching

  ch0mask  := 1 << ch0   
  ch1mask  := 1 << ch1  
  ch2mask  := 1 << ch2  
  ch3mask  := 1 << ch3  
  zcmask   := 1 << zc 
  hubch0   := @brightness  

  ' measure line frequency to determine period

  waitpne(zcmask, zcmask, 0)                                     ' wait for ZC low   
  waitpeq(zcmask, zcmask, 0)                                     ' wait for ZC high
  zctix := -cnt                                                  ' start timing
  waitpne(zcmask, zcmask, 0)                                     ' wait for ZC low          
  waitpeq(zcmask, zcmask, 0)                                     ' wait for ZC high
  zctix += cnt - 544                                             ' stop timing

  fr := clkfreq / zctix                                          ' calc 1/2-cycle frequency

  if (fr => 115)
    period := (clkfreq / 120) >> 8
    linefreq := 60
  else
    period := (clkfreq / 100) >> 8
    linefreq := 50

  cog := cognew(@dimmer, 0) + 1                                  ' start the dimmer cog
                                                                  
  return cog                                                       
                                                                  
                                                                  
pub stop                                                          
                                                                  
'' Stops AC dimming driver; frees a cog                           
                                                                  
  if (cog)                                                        
    cogstop(cog - 1)                                              
    cog := 0                                                      
                                                                  
  set_all(0)                                                      
                                                                  
                                                                  
pub set(ch, level)                                                
                                                                  
'' Sets channel to specified level                                
                                                                  
  if ((ch => 0) and (ch =< LAST_CH))                              
    brightness[ch] := 0 #> level <# 255                          ' limit to byte value
    return brightness[ch]                                        ' set channel
                                                                  
  return -1                                                      ' -1 if bad channel
                                                                  
                                                                  
pub set_all(level)                                                
                                                                  
'' Set all channels to same level                                 
                                                                  
  level := 0 #> level <# 255                                      
  bytefill(@brightness[0], level, CHANNELS)                       
                                                                  
  return level                                                    
                                                                  
                                                                  
pub high(ch)                                                      
                                                                  
'' Sets channel to on                                             
                                                                  
  if ((ch => 0) and (ch =< LAST_CH))                              
    brightness[ch] := 255                                         
    return 255                                                    
                                                                  
  return -1                                                       
                                                                  
                                                                  
pub low(ch)                                                       
                                                                  
'' Sets channel to off                                            
                                                                  
  if ((ch => 0) and (ch =< LAST_CH))                              
    brightness[ch] := 0                                           
    return 0                                                      
                                                                  
  return -1                                                       
                                                                  
                                                                  
pub toggle(ch)                                                    
                                                                  
'' Inverts channel level                                          
                                                                  
  if ((ch => 0) and (ch =< LAST_CH))                              
    if (brightness[ch] & $80)                                    ' at or above mid point
      brightness[ch] := 0                                         
      return 0                                                    
    else                                                          
      brightness[ch] := 255                                       
      return 255                                                  
                                                                  
  return -1                                                       
                                                                  
                                                                  
pub inc(ch)                                                       
                                                                  
'' Increment channel brightness                                   
                                                                  
  if ((ch => 0) and (ch =< LAST_CH))                              
    if (brightness[ch] < 255)                                     
      ++brightness[ch]                                            
    return brightness[ch]                                         
                                                                  
  return -1                                                       
                                                                  
                                                                  
pub inc_all | ch                                                  
                                                                  
'' Increment all channels                                         
                                                                  
  repeat ch from 0 to LAST_CH                                     
    if (brightness[ch] < 255)                                     
      ++brightness[ch]                                            
                                                                  
                                                                  
pub dec(ch)                                                       
                                                                  
'' Decrement channel brightness                                   
                                                                  
  if ((ch => 0) and (ch =< LAST_CH))                              
    if (brightness[ch] > 0)                                       
      --brightness[ch]                                            
    return brightness[ch]                                         
                                                                  
  return -1                                                       
                                                                  
                                                                  
pub dec_all | ch                                                  
                                                                  
'' Decrement all channels                                         
                                                                  
  repeat ch from 0 to LAST_CH                                     
    if (brightness[ch] > 0)                                       
      --brightness[ch]                                            
                                                                  
                                                                  
pub digital(value) | ch                                           
                                                                  
'' Set outputs per bits in value                                  
                                                                  
  repeat ch from 0 to LAST_CH                                     
    if (value & (1 << ch))                                       ' if bit set in value 
      set(ch, 255)                                               '  on
    else                                                         ' else
      set(ch, 0)                                                 '  off
                                                                  
                                                                  
pub masked_digital(value, mask) | ch                              
                                                                  
'' Sets enabled pins (in mask) to corresponding bit in value      
                                                                  
  repeat ch from 0 to LAST_CH                                    ' loop through outputs
    if (mask & (1 << ch))                                        ' if channel enabled
      if (value & (1 << ch))                                     ' if on
        brightness[ch] := 255                                    '  set to full 
      else                                                        
        brightness[ch] := 0                                      '  set to off
                                                                  
                                                                  
pub dcd(bit)                                                      
                                                                  
'' Sets output to decoded power-of-2 value                        
                                                                  
  if (bit < 0)                                                   ' if negative
    set_all(0)                                                   '  all off
  elseif (bit =< LAST_CH)                                        ' if valid
    digital(1 << bit)                                            '  convert to bit output
                                                                  
                                                                  
pub read(ch)                                                      
                                                                  
'' Returns channel brightness                                     
                                                                  
  if ((ch => 0) and (ch =< LAST_CH))                              
    return brightness[ch]                                         
  else                                                            
    return -1                                                     
                                                                  
                                                                  
pub address                                                       
                                                                  
'' Returns hub address of brightness array                        
                                                                  
  return @brightness


pub line_freq

'' Returns measured line frequency

  return linefreq                                             
                                                                  
                                                                  
dat                                                               
                                                                  
                        org     0                                 
                                                                  
dimmer                  mov     outa, #0                         ' clear pins
                        mov     dira, #0

                        or      dira, ch0mask                    ' set outputs
                        or      dira, ch1mask
                        or      dira, ch2mask
                        or      dira, ch3mask                        

waitzc0                 mov     t1, #DB_CYCLES                   ' wait for low
:loop                   test    zcmask, ina             wc        
        if_c            jmp     #waitzc0                          
                        djnz    t1, #:loop
                                                                  
waitzc1                 waitpeq zcmask, zcmask                   ' wait for zero cross input
                        mov     outa, #0                         ' outputs off  
                        mov     timer, period                    ' start levels timer
                        
                                                                  
getlevels               mov     t1, hubch0                       ' point to levels (in hub)
                        rdbyte  ch0level, t1                     ' read channel
                        add     ch0level, HxFFFF_FF00            ' adjust for rollover
                                                                  
                        add     t1, #1                           ' point to next element
                        rdbyte  ch1level, t1                      
                        add     ch1level, HxFFFF_FF00             
                                                                  
                        add     t1, #1                            
                        rdbyte  ch2level, t1                      
                        add     ch2level, HxFFFF_FF00             
                                                                  
                        add     t1, #1                            
                        rdbyte  ch3level, t1                      
                        add     ch3level, HxFFFF_FF00             
                                                  
                        mov     idx, #255
                        add     timer, cnt                           
                                                                  
dimloop                 add     ch0level, #1            wc       ' inc value
        if_c            or      outa, ch0mask                    ' turn on, leave on  
                                                                  
                        add     ch1level, #1            wc        
        if_c            or      outa, ch1mask                     
                                                                  
                        add     ch2level, #1            wc        
        if_c            or      outa, ch2mask                     
                                                                  
                        add     ch3level, #1            wc        
        if_c            or      outa, ch3mask                     
                                                                  
                        waitcnt timer, period                    ' let this period finish
                        djnz    idx, #dimloop                     
                                                                  
                        jmp     #waitzc1                         ' next cycle
                                                                  
' --------------------------------------------------------------------------------------------------

HxFFFF_FF00             long    $FFFF_FF00

ch0mask                 long    0-0                              ' pin mask for channel 0 
ch1mask                 long    0-0                              ' pin mask for channel 1
ch2mask                 long    0-0                              ' pin mask for channel 2
ch3mask                 long    0-0                              ' pin mask for channel 3
zcmask                  long    0-0                              ' pin mask for zero cross
hubch0                  long    0-0                              ' hub address of brightness array
period                  long    0-0                              ' 1/256th of half-cycle

timer                   res     1                                ' for period timing
                                                                  
ch0level                res     1                                ' channel levels
ch1level                res     1                                 
ch2level                res     1                                 
ch3level                res     1                                 
                                                                  
idx                     res     1                                 
                                                                  
t1                      res     1                                 
t2                      res     1                                 
                                                                  
                        fit     496                                

                        
dat { license } 

{{

  Copyright (c) 2010-2015 Jon McPhalen  

  MIT License

  Permission is hereby granted, free of charge, to any person obtaining a copy of this
  software and associated documentation files (the "Software"), to deal in the Software
  without restriction, including without limitation the rights to use, copy, modify,
  merge, publish, distribute, sublicense, and/or sell copies of the Software, and to
  permit persons to whom the Software is furnished to do so, subject to the following
  conditions:

  The above copyright notice and this permission notice shall be included in all copies
  or substantial portions of the Software.

  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED,
  INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A
  PARTICULAR PURPOSE AND NON-INFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
  HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF
  CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE
  OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

}}  